# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Custom exceptions raised during sanity checks for itwinai."""


class SanityCheckError(Exception):
    """Base exception for all sanity check errors."""
