# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from .dummy_components import (
    FakeGetter,
    FakeGetterExec,
    FakePreproc,
    FakePreprocExec,
    FakeSaver,
    FakeSaverExec,
    FakeSplitter,
    FakeSplitterExec,
    FakeTrainer,
    FakeTrainerExec,
)

_ = (
    FakeGetter,
    FakeGetterExec,
    FakePreproc,
    FakePreprocExec,
    FakeSaver,
    FakeSaverExec,
    FakeSplitter,
    FakeSplitterExec,
    FakeTrainer,
    FakeTrainerExec,
)
